/* -*- mode: c++; c-basic-offset: 3; -*- */
#ifndef DQ_MODULE_HH
#define DQ_MODULE_HH 1

#include "MultiStream.hh"
#include "DQ_Bit.hh"
#include "frame_name.hh"
#include "FrWriter.hh"
#include "TSeries.hh"
#include "FilterBase.hh"
#include <vector>
#include <string>
#include <iosfwd>

/**  The %DQ_Module class implements the %DQ_Module monitor. This montor calls 
  *  a boolean-valued c++ function of one or more time series and uses the 
  *  result value to set a specified bit in a uint32_t bit-mask channel. The 
  *  monitor loads a plugin.
  */
class DQ_Module : public MultiStream {
public:
   class dq_bitmask;
   class filter_template;
   class process_channel;
public:
   /**  %DQ_Module monitor constructor. The monitor requires a single json
     *  configuration file path, and also parses the usual MultiStream 
     *  command line arguments.
     */
   DQ_Module(int argc, const char* argv[]);

   /**  %DQ_Module destructor.
     */
   ~DQ_Module(void);

   /** discard write buffer data up to the specified time.
    */
   void align_buffer(const Time& t);
   
   /**  Collect all the channel names requested as input to the dq bit 
     *  evaluation functions and the channels requested to be copied to
     *  the output frame.
     */ 
   void book_channels(void);

   /**  Find a TSeries in the processed channel list, the dq vector channel 
     *  list, or raw-data channel and return the location: 0 if not found or
     *  1 ... n for other sources.
     *  \brief Locate an operand channel
     *  \param chan channel name string.
     *  \return channel type 0 - n.
     */
   int chan_type(const std::string& chan) const;

   /**  Find a TSeries in the Processed channel list, the dq vector channel 
     *  list, or raw-data channel. An empty time series reference is returned 
     *  if the requested channel is not found.
     *  \brief Find an operand channel
     *  \param chan channel name string.
     *  \return constant reference to a channel.
     */
   const TSeries& find(const std::string& chan) const;
   
   /**  Call the init methods for each of the defined flag instances.
     *  \brief Initialize flag evaluation.
     */
   void init_all(void);

   /**  Create output channels for all defined flags. This method also 
     *  builds a mask for each bit that will be used.
     *  \brief Initialize output channels.
     */
   void init_output_channels(void);
   
   /**  Parse the json configuration file.
     */
   void config(const std::string& cfg_file);

   /**  Process a stride of input data.
     */
   void ProcessData(void);
   void put_config(std::ostream& out) const;
   void set_flag(DQ_bit* dq_def, const Time& t0, Interval dt);
   void write_frame(const Time& t0, Interval dT);
 private:
   double stride;            ///<  computation stride (frame length)
   double flag_step;         ///<  mask calculation stride length
   int    parity_bit;        ///< Parity bit number or no parity if \< 0
   bool   invert;
   frame_name frame_path;
   std::string compression;
   std::string config_file;
   std::string config_version;
   uint32_t    config_crc32;
   //std::vector<std::string> copy_channels;
   std::vector<std::string> copy_history;
   std::vector<DQ_bit*> flag_vect;
   FrWriter fr_writer;
   Time     output_start;
   Time     current;
   Time     write_time;

   typedef std::map<std::string, dq_bitmask> flag_map_type;
   typedef flag_map_type::iterator flag_map_iter;
   typedef flag_map_type::const_iterator const_flag_map_iter;
   flag_map_type flag_map;
   size_t   frames_out;
   
   typedef std::map<std::string, filter_template> filter_map;
   typedef filter_map::value_type filter_map_node;
   filter_map filterDB;

   typedef std::vector<process_channel> processed_list;
   processed_list     proc_list;
};

//======================================  define the flag map data type
class DQ_Module::dq_bitmask {
public:
   dq_bitmask(void);
   ~dq_bitmask();

   /**  Add the numbered bit to the enable mask (i.e. the mask of all
    *  bits set by the defined dq flags.
    */
   void enable(int bitno);

   /**  Preset the bit time series to all trigger bits OK. The evaluated 
    *  flags are then "OR"ed with the bit-masks.
    *  \brief Preset the time series.
    *  \param t0 Start time for the current stride.
    *  \param dT flag sample interval
    *  \param N  Number of time-series samples in this stride.
    */
   void preset(const Time& t0, Interval dT, size_t N);

   /**  Return a constant reference to the time series for this channel.
    *  \brief DQ channel time series reference.
    */
   const TSeries& ref_series(void) const {return _series;}

   /**  Return a constant reference to the time series for this channel.
    *  \brief DQ channel time series reference.
    */
   TSeries& ref_series(void) {return _series;}
private:
   TSeries  _series;
   uint32_t _enable_mask;
};

//======================================  Internal filter database class
class DQ_Module::filter_template {
public:
   enum filter_enum {kf_none, kf_pipe, kf_ypipe};
   filter_template(void);
   filter_template(const std::string& name, const FilterBase& base,
		   filter_enum type);
   filter_template(const filter_template& ft);
   filter_template& operator=(const filter_template& ft);
   ~filter_template(void);
   TSeries apply(const TSeries& x);
   TSeries apply(const TSeries& x, const TSeries& y);
   const FilterBase* get(void) const;
   FilterBase* get(void);
   const std::string& name(void) const;
   void reset(void);
   std::string type(void) const;
private:
   std::string _name;
   std::unique_ptr<FilterBase> _filter;
   filter_enum _type;
};

//=======================================  Processed channel class
/**  Process channel control block.
 */
class DQ_Module::process_channel {
public:
   /** Enumerate the processing modes.
     */
   enum process_enum {pk_undefined,
		      pk_none,
		      pk_nowrite,
		      pk_copy,
		      pk_write};  
public:
   /**  Construct a processed channel status class instance.
     *  \brief Processed chanel status class constructor.
     *  \param chan channel name.
     */
   process_channel(const std::string& chan);

   /**  Destroy a processed channel status class instance.
     *  \brief Processed chanel status class destructor.   
     */
   ~process_channel(void);

   void apply(const TSeries& x);
   void apply(const TSeries& x, const TSeries& y);

   /**  Buffer the time series for writing.
     */
   TSeries& buffer(void);
   const std::string& channel_x(void) const;
   const std::string& channel_y(void) const;
   void dump(std::ostream&  out) const;
   const filter_template& filter(void) const;
   filter_template& filter(void);
   void flush(void);
   const std::string& name(void) const;
   process_enum process_type(void) const;
   const TSeries& ref_series(void) const;
   TSeries& ref_series(void);
   void reset(void);
   void set_channel_x(const std::string& xname);
   void set_channel_y(const std::string& yname);
   void set_filter(const filter_template& filt);
   void set_process_type(process_enum mode);
   void set_process_type(const std::string& mode);
   std::string type(void) const;
   bool write(void) const;
private:
   process_enum    _process_type;
   std::string     _chan_name;
   std::string     _chan_inx;
   std::string     _chan_iny;
   filter_template _filt;
   TSeries         _output;
   TSeries         _write_buffer;
};


//======================================  Inline methods

//======================================  dq_bitmask
inline void
DQ_Module::dq_bitmask::enable(int bitno) {
   _enable_mask |= (1<<bitno);
}

//====================================== filter_template
inline const FilterBase*
DQ_Module::filter_template::get(void) const {
   return _filter.get();
}

inline FilterBase*
DQ_Module::filter_template::get(void) {
   return _filter.get();
}

inline const std::string&
DQ_Module::filter_template::name(void) const {
   return _name;
}

//======================================  process_channel
inline const std::string& 
DQ_Module::process_channel::channel_x(void) const {
   return _chan_inx;
}

inline const std::string&
DQ_Module::process_channel::channel_y(void) const {
   return _chan_iny;
}

inline const DQ_Module::filter_template&
DQ_Module::process_channel::filter(void) const {
   return _filt;
}

inline DQ_Module::filter_template&
DQ_Module::process_channel::filter(void) {
   return _filt;
}

inline const std::string&
DQ_Module::process_channel::name(void) const {
   return _chan_name;
}

inline DQ_Module::process_channel::process_enum
DQ_Module::process_channel::process_type(void) const {
   return _process_type;
}

inline const TSeries&
DQ_Module::process_channel::ref_series(void) const {
   return _output;
}

inline TSeries&
DQ_Module::process_channel::ref_series(void) {
   return _output;
}

inline bool 
DQ_Module::process_channel::write(void) const {
   switch (_process_type) {
   case pk_copy:
   case pk_write:
      return true;
   default:
      return false;
   }
}

#endif
